int SOCK_STREAM 	= 1,
    SOCK_DGRAM		= 2,
    SOCK_RAW		= 3,
    SOCK_RDM		= 4,
    SOCK_SEQPACKET	= 5;

int AF_UNSPEC		= 0,
    AF_UNIX		= 1,
    AF_INET		= 2,
    AF_IMPLINK		= 3,
    AF_PUP		= 4,
    AF_CHAOS		= 5,
    AF_NS		= 6,
    AF_ISO		= 7,
    AF_OSI		= AF_ISO,
    AF_ECMA		= 8,
    AF_DATAKIT		= 9,
    AF_CCITT		= 10,
    AF_SNA		= 11,
    AF_DECnet		= 12,
    AF_DLI		= 13,
    AF_LAT		= 14,
    AF_HYLINK		= 15,
    AF_APPLETALK	= 16,
    AF_ROUTE		= 17,
    AF_LINK		= 18,
    pseudo_AF_XTP	= 19,

    AF_MAX		= 20;

int SOMAXCONN		= 5;

type IOVec_t = struct {
    arbptr iov_base;
    ulong iov_len;
};

type MsgHdr_t = struct {
    arbptr msg_name;
    ulong msg_nameLen;
    *IOVec_t msg_iov;
    ulong msg_iovLen;
    arbptr msg_control;
    ulong msg_controlLen;
    long msg_flags;
};

long
    MSG_OOB		= 0x1,
    MSG_PEEK		= 0x2,
    MSG_DONTROUTE	= 0x4,
    MSG_EOR		= 0x8,
    MSG_TRUNC		= 0x10,
    MSG_CTRUNC		= 0x20,
    MSG_WAITALL 	= 0x40;

type CMsgHdr_t = struct {
    ulong cmsg_len;
    long cmsg_level;
    long cmsg_type;
};

type INAddr_t = struct {
    ulong s_addr;
};

type SockAddrIn_t = struct {
    ushort sin_len;
    ushort sin_family;
    uint sin_port;
    INAddr_t sin_addr;
    [8] byte sin_zero;
};

type IPOpts_t = struct {
    INAddr_t ip_dst;
    [40] byte ip_opts;
};

ulong FD_SETSIZE = 64;

ulong NFDBITS = FD_SETSIZE / (8 * sizeof(ulong));

type FDSet_t = struct {
    [NFDBITS] ulong fds_bits;
};

type TimeVal_t = struct {
    ulong tv_sec;
    ulong tv_usec;
};

type HostEnt_t = struct {
    *char h_name;
    **char h_aliases;
    long h_addrtype;
    long h_length;
    **char h_addrlist;
};

type NetEnt_t = struct {
    *char n_name;
    **char n_aliases;
    long n_addrtype;
    ulong n_net;
};

type ServEnt_t = struct {
    *char s_name;
    **char s_aliases;
    long s_port;
    *char s_proto;
};

type ProtoEnt_t = struct {
    *char p_name;
    **char p_aliases;
    long p_proto;
};

/* return values for netdb routines */

ulong
    HOST_NOT_FOUND	= 1,
    TRY_AGAIN		= 2,
    NO_RECOVERY 	= 3,
    NO_DATA		= 4,
    NO_ADDRESS		= NO_DATA;

/* priorities for Syslog */

ulong
    LOG_EMERG	= 0,	/* system is unusable */
    LOG_ALERT	= 1,	/* action must be taken immediately */
    LOG_CRIT	= 2,	/* critical conditions */
    LOG_ERR	= 3,	/* error conditions */
    LOG_WARNING = 4,	/* warning conditions */
    LOG_NOTICE	= 5,	/* normal but significant condition */
    LOG_INFO	= 6,	/* informational */
    LOG_DEBUG	= 7;	/* debug-level messages */

extern
    OpenBsdsocketLibrary(ulong version)*SocketBase_t,
    CloseBsdsocketLibrary()void,

    Socket(ulong domain, typ, protocol)long,
    Bind(long socket; *SockAddrIn_t name; long nameLen)long,
    Listen(long socket, count)long,
    Accept(long socket; *SockAddrIn_t fromAddr; *ulong fromLen)long,
    Connect(long socket; *SockAddrIn_t addr; ulong nameLen)long,
    Send(long socket; arbptr msg; ulong len, flags)long,
    SendTo(long socket; arbptr msg; ulong len, flags; *SockAddrIn_t addr;
	ulong toLen)long,
    SendMsg(long socket; *MsgHdr_t msg; ulong flags)long,
    Recv(long socket; arbptr msg; ulong len, flags)long,
    RecvFrom(long socket; arbptr buf; ulong len, flags; *SockAddrIn_t addr;
	*ulong fromLen)long,
    RecvMsg(long socket; *MsgHdr_t msg; ulong flags)long,
    Shutdown(long socket; long how)long,
    SetSockOpt(long socket; ulong level, optName;
	arbptr optVal; ulong optLen)long,
    GetSockOpt(long socket; ulong level, optName;
	arbptr optVal; *ulong optLen)long,
    GetSockName(long socket; *SockAddrIn_t name; *ulong nameLen)long,
    GetPeerName(long socket; *SockAddrIn_t name; *ulong nameLen)long,

    IoctlSocket(long socket; ulong request; arbptr argp)long,
    CloseSocket(long socket)long,
    WaitSelect(ulong nFd; *FDSet_t readFds, writeFds, exceptFds;
	*TimeVal_t timeVal; *ulong maskP)long,

    Dup2Socket(long fd1, fd2)long,

    GetDTableSize()ulong,
    SetSocketSignals(ulong SIGINTR, SIGIO, SIGURG)void,
    SetErrnoPtr(arbptr errno_p; ulong size)long,

    GetSocketEvents(*ulong eventMaskP)long,

    Errno()long,

    GetHostName(*char hostName; ulong size)long,
    GetHostId()long,

    ObtainSocket(long id, domain, typ, protocol)long,
    ReleaseSocket(long socket; ulong id)long,
    ReleaseCopyOfSocket(long socket; ulong id)long,

    Inet_Addr(*byte x)ulong,
    Inet_Network(*byte x)ulong,
    Inet_NtoA(ulong s_addr)*char,
    Inet_MakeAddr(ulong net, lna)ulong,
    Inet_LnaOf(long s_addr)ulong,
    Inet_NetOf(long s_addr)ulong,

    GetHostByName(*char name)*HostEnt_t,
    GetHostByAddr(arbptr addr; ulong len, typ)*HostEnt_t,
    GetNetByName(*char name)*NetEnt_t,
    GetNetByAddr(long net, typ)*NetEnt_t,
    GetServByName(*char name, proto)*ServEnt_t,
    GetServByPort(ulong port; *char proto)*ServEnt_t,
    GetProtoByName(*char name)*ProtoEnt_t,
    GetProtoByNumber(long proto)*ProtoEnt_t,

    SyslogA(ulong pri; *char fmt; *arbptr other)void;
