#drinc:exec/nodes.g
#drinc:exec/lists.g
#drinc:exec/memory.g
#drinc:exec/tasks.g
#drinc:exec/ports.g
#drinc:exec/io.g

bool PARANOID = true ;

proc NewList(register *List_t l)void:

    l*.lh_Head := pretend(&l*.lh_Tail, *Node_t);
    l*.lh_TailPred := pretend(&l*.lh_Head, *Node_t);
    l*.lh_Tail := nil;
corp;

proc BeginIO(*IORequest_t io)void:
    uint
	OP_MOVEL = 0x2000,
	OP_JSR	 = 0x4e80,
	M_ADIR = 0o1,
	M_DISP = 0o5,
	R_SP = 7,
	R_A1 = 1,
	R_A5 = 5,
	R_A6 = 6,
	R_FP = 6;

    int
	DEVICE_OFFSET = sizeof(Message_t),
	BEGINIO_OFFSET = -30;
    ulong saveA1, saveA5, saveA6;

    code (
	/* save A1 in saveA1 */
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A1,
	saveA1,
	/* save A5 in saveA5 */
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A5,
	saveA5,
	/* copy FP into A5 */
	OP_MOVEL | R_A5 << 9 | M_ADIR << 6 | M_ADIR << 3 | R_FP,
	/* get IORequest address into A1 */
	OP_MOVEL | R_A1 << 9 | M_ADIR << 6 | M_DISP << 3 | R_FP,
	io,
	/* get Device pointer from io into A6 */
	OP_MOVEL | R_A6 << 9 | M_ADIR << 6 | M_DISP << 3 | R_A1,
	DEVICE_OFFSET,
	/* call the BEGINIO function in the device */
	OP_JSR | M_DISP << 3 | R_A6,
	BEGINIO_OFFSET,
	/* restore FP from A5 */
	OP_MOVEL | R_FP << 9 | M_ADIR << 6 | M_ADIR << 3 | R_A5,
	/* restore A5 */
	OP_MOVEL | R_A5 << 9 | M_ADIR << 6 | M_DISP << 3 | R_FP,
	saveA5,
	/* restore A1 */
	OP_MOVEL | R_A1 << 9 | M_ADIR << 6 | M_DISP << 3 | R_FP,
	saveA1
    );
corp;

proc CreatePort(*char name; short pri)*MsgPort_t:
    register *MsgPort_t port;
    long sigBit;

    sigBit := AllocSignal(-1);
    if sigBit = -1 then
	nil
    else
	port := AllocMem(sizeof(MsgPort_t), MEMF_CLEAR | MEMF_PUBLIC);
	if port = nil then
	    FreeSignal(sigBit);
	    nil
	else
	    port*.mp_Node.ln_Name := name;
	    port*.mp_Node.ln_Pri := pri;
	    port*.mp_Node.ln_Type := NT_MSGPORT;

	    port*.mp_Flags := PA_SIGNAL;
	    port*.mp_SigBit := sigBit;
	    port*.mp_SigTask := FindTask(nil);

	    if name ~= nil then
		AddPort(port);
	    else
		NewList(&port*.mp_MsgList);
	    fi;

	    port
	fi
    fi
corp;

proc DeletePort(*MsgPort_t port)void:

    if port*.mp_Node.ln_Name ~= nil then
	RemPort(port);
    fi;
    if PARANOID then
	/* Why do they want to do this? Is there some convention of flushing
	   lists and stuff that I don't know about? */
	/* One use is that two programs communicating over a message port
	   can sometimes survive if one abruptly dies. Any use of a port to
	   another program should check for the -1 in lh_Head, and abort
	   itself if it finds it. */
	port*.mp_Node.ln_Type := 0xff;
	port*.mp_MsgList.lh_Head := pretend(-1, *Node_t);
    fi;
    FreeSignal(port*.mp_SigBit);
    FreeMem(port, sizeof(MsgPort_t));
corp;

proc CreateStdIO(*MsgPort_t ioReplyPort)*IOStdReq_t:
    register *IOStdReq_t ioStdReq;

    if ioReplyPort = nil then
	nil
    else
	ioStdReq := AllocMem(sizeof(IOStdReq_t), MEMF_CLEAR | MEMF_PUBLIC);
	if ioStdReq = nil then
	    nil
	else
	    ioStdReq*.io_io.io_Message.mn_Node.ln_Type := NT_MESSAGE;
/*
	    ioStdReq*.io_io.io_Message.mn_Node.ln_Pri := 0;
*/
	    ioStdReq*.io_io.io_Message.mn_ReplyPort := ioReplyPort;
	    ioStdReq*.io_io.io_Message.mn_Length :=
		sizeof(IOStdReq_t) - sizeof(Message_t);
	    ioStdReq
	fi
    fi
corp;

proc DeleteStdIO(*IOStdReq_t ioStdReq)void:

    if PARANOID then
	/* again, why are we doing this? */
	ioStdReq*.io_io.io_Message.mn_Node.ln_Type := 0xff;
	ioStdReq*.io_io.io_Device := pretend(-1, *Device_t);
	ioStdReq*.io_io.io_Unit := pretend(-1, *Unit_t);
    fi;
    FreeMem(ioStdReq, sizeof(IOStdReq_t));
corp;

proc CreateTask(*char name; short pri; *byte initPC; ulong stackSize)*Task_t:
    register *Task_t newTask;
    ulong dataSize;

    dataSize := (stackSize & 0xfffffc) + 1;
    newTask := AllocMem(sizeof(Task_t) + dataSize, MEMF_CLEAR | MEMF_PUBLIC);
    if newTask = nil then
	nil
    else
	newTask*.tc_SPLower := pretend(newTask, *byte) + sizeof(Task_t);
	newTask*.tc_SPUpper :=
	   pretend((pretend(newTask*.tc_SPLower, ulong) + dataSize) & 0xfffffe,
		   *byte);
	newTask*.tc_SPReg := newTask*.tc_SPUpper;
	newTask*.tc_Node.ln_Type := NT_TASK;
	newTask*.tc_Node.ln_Pri := pri;
	newTask*.tc_Node.ln_Name := name;
	AddTask(newTask, initPC, nil);
	newTask
    fi
corp;

proc DeleteTask(*Task_t tc)void:

    RemTask(tc);
    FreeMem(tc, tc*.tc_SPUpper - pretend(tc, *byte) + 1);
corp;

proc CreateExtIO(*MsgPort_t ioReplyPort; ulong size)*IORequest_t:
    register *IORequest_t ioReq;

    if ioReplyPort = nil then
	nil
    else
	ioReq := AllocMem(size, MEMF_CLEAR | MEMF_PUBLIC);
	if ioReq = nil then
	    nil
	else
	    ioReq*.io_Message.mn_Node.ln_Type := NT_MESSAGE;
/*
	    ioReq*.io_Message.mn_Node.ln_Pri := 0;
*/
	    ioReq*.io_Message.mn_ReplyPort := ioReplyPort;
	    ioReq*.io_Message.mn_Length := size - sizeof(Message_t);
	    ioReq
	fi
    fi
corp;

proc DeleteExtIO(*IORequest_t ioExt; ulong size)void:

    if PARANOID then
	ioExt*.io_Message.mn_Node.ln_Type := 0xff;
	ioExt*.io_Device := pretend(-1, *Device_t);
	ioExt*.io_Unit := pretend(-1, *Unit_t);
    fi;
    FreeMem(ioExt, size);
corp;
