#drinc:exec/miscellaneous.g
#drinc:exec/nodes.g
#drinc:exec/libraries.g
#drinc:exec/devices.g
#drinc:exec/io.g
#drinc:exec/resident.g
#drinc:exec/memory.g
#drinc:libraries/dos.g
#drinc:util.g

/*
 * FDCompile.d - convert '.fd' files to interface libraries for them.
 *
 *	Author: Chris Gray
 *	Date: June 21, 1986
 */

bool DEBUG = false;

uint
    BUFF_SIZE = 10000,			/* make it big, minimize I/O */
    NAME_SIZE = 100;			/* maximum symbol name */

/* Some machine and operating system definitions: */

uint
    R_LIB	= 6,
    R_SP	= 7,

    M_DDIR	= 0o0,
    M_ADIR	= 0o1,
    M_INDIR	= 0o2,
    M_INC	= 0o3,
    M_DEC	= 0o4,
    M_DISP	= 0o5,
    M_SPECIAL	= 0o7,

    M_ABSLONG	= 0o1,
    M_PCDISP	= 0o2,
    M_IMM	= 0o4,

    OP_BTST	= 0x0800,
    OP_BCLR	= 0x0880,
    OP_BSET	= 0x08c0,
    OP_MOVEL	= 0x2000,
    OP_MOVEW	= 0x3000,
    OP_LEA	= 0x41c0,
    OP_CLR	= 0x4200,
    OP_MOVEML	= 0x48c0,
	MOVEM_SAVE	= 0x0000,
	MOVEM_RESTORE	= 0x0400,
    OP_TST	= 0x4a00,
    OP_JSR	= 0x4e80,
    OP_JMP	= 0x4ec0,
    OP_RTS	= 0x4e75,
    OP_ADDQ	= 0x5000,
    OP_SUBQ	= 0x5100,
    OP_BCC	= 0x6000,
	CC_T		= 0x0,
	CC_EQ		= 0x7,
	CC_NE		= 0x6,
    OP_BSR	= 0x6100,
    OP_SUB	= 0x9000,
    OP_ADD	= 0xd000,

    S_BYTE	= 0,
    S_WORD	= 1,
    S_LONG	= 2,

    OM_TO_REG	= 0b000,
    OM_TO_MEM	= 0b100,

    OM_WORD_ADDR= 0b011,
    OM_LONG_ADDR= 0b111;

ulong ABS_EXEC_BASE = 4;
int
    LVO_FREE_MEM      = -30 - 6 * 30,
    LVO_REMOVE	      = -30 - 6 * 37,
    LVO_OPEN_LIBRARY  = -30 - 6 * 87,
    LVO_CLOSE_LIBRARY = -30 - 6 * 64,
    LVO_OPEN_DEVICE   = -30 - 6 * 69,
    LVO_CLOSE_DEVICE  = -30 - 6 * 70;

ulong
    HUNK_UNIT	= 999,
    HUNK_CODE	= 1001,
    HUNK_BSS	= 1003,
    HUNK_RELOC32= 1004,
    HUNK_EXT	= 1007,
    HUNK_SYMBOL = 1008,
    HUNK_END	= 1010;

byte
    EXT_DEF	= 1,
    EXT_ABS	= 2,
    EXT_REF32	= 129;

type
    Output_t = struct {
	[BUFF_SIZE] byte o_buffer;
	Handle_t o_fd;
	unsigned BUFF_SIZE o_pos;
    },

    FuncList_t = struct {
	*FuncList_t fl_next;
	[NAME_SIZE] char fl_name;
    };

[BUFF_SIZE] char FdBuffer;		/* buffer for input .fd file */
Handle_t
    Stdout,				/* the standard output */
    FdFd;				/* the input file */
unsigned BUFF_SIZE
    FdPos,				/* position in input buffer */
    FdMax;				/* amount in input buffer */

Output_t
    LibOutput,
    LVOOutput,
    SymOutput,
    HeadOutput;

uint CodePos;				/* code position counter */

uint CurrentLine;			/* current line in input */
*char CurrentFile;			/* name of current file */

char
    InputChar,				/* current input character */
    Version1,				/* first digit of library version */
    Version2;				/* second digit of library version */

[NAME_SIZE] char
    InputName,				/* current input name */
    EntryName,				/* saved name of current entry */
    BaseName;				/* saved name of library base */

bool
    Eof,				/* true -> end-of-file on input */
    PassOne,				/* pass one of code gen */
    IsDevice,				/* true for device, not library */
    MakeLib,				/* produce lib output file */
    PrintLib,				/* print LVO offsets */
    MakeLVO,				/* produce LVO output file */
    MakeSym,				/* produce sym output file */
    PrintSym,				/* print sym values */
    PrintAll,				/* also print reserved slots */
    MakeHead;				/* build disk library header */

*Library_t LibPtr;			/* pointer to the opened library */
*IORequest_t Req;			/* IO request for open device */

*FuncList_t FuncList;			/* list of functions */
**FuncList_t FuncTail;			/* build in forward order */
uint FuncCount;

/*
 * writeString -
 *	Write a string message to the standard output.
 */

proc writeString(*char message)void:
    register *char p;

    p := message;
    while p* ~= '\e' do
	p := p + 1;
    od;
    ignore Write(Stdout, message, p - message);
corp;

/*
 * writeNumber -
 *	Write an unsigned number to the standard output.
 */

proc writeNumber(register uint n)void:
    [6] char buffer;
    register *char p;

    p := &buffer[5];
    p* := '\e';
    while
	p := p - 1;
	p* := n % 10 + '0';
	n := n / 10;
	n ~= 0
    do
    od;
    writeString(p);
corp;

/*
 * writeHex - write a 16 bit hex value to the standard output.
 */

proc writeHex(register uint n)void:
    [5] char buf;
    register uint digit, i;
    register *char p;

    p := &buf[4];
    p* := '\e';
    for i from 3 downto 0 do
	digit := n & 0xf;
	p := p - sizeof(char);
	p* := if digit >= 10 then digit - 10 + 'a' else digit + '0' fi;
	n := n >> 4;
    od;
    writeString(&buf[0]);
corp;

/*
 * closeOne - flush and close an output unit.
 */

proc closeOne(*Output_t o)void:
    extern flush(*Output_t o)void;

    if o*.o_fd ~= 0 then
	flush(o);
	Close(o*.o_fd);
    fi;
corp;

/*
 * closeAll - we are done, close all files.
 */

proc closeAll()void:
    register *FuncList_t fl;

    closeOne(&HeadOutput);
    closeOne(&SymOutput);
    closeOne(&LVOOutput);
    closeOne(&LibOutput);
    Close(FdFd);
    while FuncList ~= nil do
	fl := FuncList;
	FuncList := fl*.fl_next;
	FreeMem(fl, sizeof(FuncList_t));
    od;
corp;

/*
 * abort -
 *	Abort, printing a message, closing files, and returning the
 *	given status code.
 */

proc abort(*char message; uint status)void:

    writeString(message);
    closeAll();
    Exit(status);
corp;

/*
 * flush - flush one of the outputs.
 */

proc flush(register *Output_t o)void:

    if o*.o_pos ~= 0 then
	if Write(o*.o_fd, &o*.o_buffer[0], o*.o_pos) ~= o*.o_pos then
	    o*.o_pos := 0;
	    abort("Write to output failed.\n", 30);
	fi;
	o*.o_pos := 0;
    fi;
corp;

/*
 * addByte - add a byte to one of the outputs.
 */

proc addByte(register *Output_t o; byte b)void:

    if o ~= &LibOutput or not PassOne then
	if o*.o_pos = BUFF_SIZE then
	    flush(o);
	fi;
	o*.o_buffer[o*.o_pos] := b;
	o*.o_pos := o*.o_pos + 1;
    fi;
    if o = &LibOutput then
	CodePos := CodePos + 1;
    fi;
corp;

/*
 * addWord - add a word to one of the outputs.
 */

proc addWord(*Output_t o; uint w)void:

    addByte(o, w >> 8);
    addByte(o, w);
corp;

/*
 * addLong - add a long to one of the outputs.
 */

proc addLong(*Output_t o; ulong l)void:

    addWord(o, l >> 16);
    addWord(o, l);
corp;

/*
 * addName - add the given name, with the given high byte on its length.
 */

proc addName(register *Output_t o; register *char name; byte kind)void:
    register *char p;
    register ulong length;

    p := name;
    length := 0;
    while p* ~= '\e' do
	p := p + sizeof(char);
	length := length + 1;
    od;
    length := (length + 3) / 4 * 4;
    addLong(o, make(kind, ulong) << 24 | (length / 4));
    while name* ~= '\e' do
	addByte(o, name* - '\e');
	name := name + sizeof(char);
	length := length - 1;
    od;
    while length ~= 0 do
	addByte(o, 0);
	length := length - 1;
    od;
corp;

/*
 * addString - add a simple string to the output unit.
 */

proc addString(register *Output_t o; register *char st)void:

    while st* ~= '\e' do
	addByte(o, st* - '\e');
	st := st + sizeof(char);
    od;
corp;

/*
 * syntax -
 *	A syntax error in the .fd file. Complain and abort.
 */

proc syntax(*char message)void:

    writeString(CurrentFile);
    writeString(", ");
    writeNumber(CurrentLine);
    writeString(": ");
    abort(message, 20);
corp;

/*
 * getChar -
 *	Get the next input character into 'InputChar'.
 */

proc getChar()void:

    if Eof then
	syntax("Unexpected end of file on input.\n");
    fi;
    if FdPos = FdMax then
	if DEBUG then
	    writeString("reading a bufferful\n");
	fi;
	FdPos := 0;
	FdMax := Read(FdFd, &FdBuffer[0], BUFF_SIZE);
	if FdMax = 0 then
	    Eof := true;
	fi;
    fi;
    InputChar := FdBuffer[FdPos];
    FdPos := FdPos + 1;
corp;

/*
 * getName -
 *	Get a name or word into 'InputName'.
 */

proc getName()void:
    register uint i;

    i := 0;
    while InputChar ~= ' ' and InputChar ~= '\n' and InputChar ~= '(' and
	InputChar ~= ',' and InputChar ~= ')' and InputChar ~= '.'
    do
	InputName[i] := InputChar;
	getChar();
	if Eof then
	    syntax("No terminator for name\n");
	fi;
	i := i + 1;
    od;
    InputName[i] := '\e';
corp;

/*
 * buildName -
 *	Build a name in a buffer from 3 parts. Return the total length.
 */

proc buildName(register *char buffer, p1, p2, p3)void:
    register uint length;

    while p1* ~= '\e' do
	buffer* := p1*;
	buffer := buffer + sizeof(char);
	p1 := p1 + sizeof(char);
    od;
    while p2* ~= '\e' do
	buffer* := p2*;
	buffer := buffer + sizeof(char);
	p2 := p2 + sizeof(char);
    od;
    while p3* ~= '\e' do
	buffer* := p3*;
	buffer := buffer + sizeof(char);
	p3 := p3 + sizeof(char);
    od;
    buffer* := '\e';
corp;

/*
 * produceSetDevice -
 *	Produce the SetXXXDevice routine.
 */

proc produceSetDevice(*char deviceName)void:
    [NAME_SIZE] char entryName;
    register *Output_t lo;

    if DEBUG then
	writeString("produceSetDevice\n");
    fi;
    lo := &LibOutput;
    /* first, add a bss hunk with space for the device base pointer */
    addLong(lo, HUNK_UNIT);
    addLong(lo, 0);
    addLong(lo, HUNK_BSS);
    addLong(lo, 4 / 4);
    /* add an EXT hunk to define the base name as in the bss hunk */
    addLong(lo, HUNK_EXT);
    addName(lo, &BaseName[0], EXT_DEF);
    addLong(lo, 0);
    addLong(lo, 0);
    addLong(lo, HUNK_END);
    /* chuck out the code for the SetXXXDevice routine */
    addLong(lo, HUNK_UNIT);
    addLong(lo, 0);
    addLong(lo, HUNK_CODE);
    addLong(lo, (10 + 2) / 4);
    addWord(lo, OP_MOVEL | 1 << 9 | M_ADIR << 6 | M_INC << 3 | R_SP);
    addWord(lo, OP_MOVEL | M_ABSLONG << 9 | M_SPECIAL << 6 | M_INC<<3 | R_SP);
    addLong(lo, 0);	/* reference the above base variable */
    addWord(lo, OP_JMP | M_INDIR << 3 | 1);
    /* pad to a long word boundary */
    addWord(lo, 0);
    /* capitalize the first letter of the device name */
    if deviceName* >= 'a' and deviceName* <= 'z' then
	deviceName* := deviceName* - 32;
    fi;
    /* build the name of the Set routine */
    buildName(&entryName[0], "Set", deviceName, "Device");
    /* define the entry point and reference the base variable */
    addLong(lo, HUNK_EXT);
    addName(lo, &entryName[0], EXT_DEF);
    addLong(lo, 0);
    addName(lo, &BaseName[0], EXT_REF32);
    addLong(lo, 1);
    addLong(lo, 4);
    addLong(lo, 0);
    addLong(lo, HUNK_END);
    if DEBUG then
	writeString("leaving produceSetDevice\n");
    fi;
corp;

/*
 * produceOpenClose -
 *	Produce the OpenXXXLibrary and CloseXXXLibrary routines.
 */

proc produceOpenClose(*char libraryName)void:
    [NAME_SIZE] char entryName;
    register *Output_t lo;
    register *char p;
    register uint libraryNameLength;

    if DEBUG then
	writeString("produceOpenClose\n");
    fi;
    lo := &LibOutput;
    /* find out how long the library name is */
    p := libraryName;
    libraryNameLength := 9;
    while p* ~= '\e' do
	p := p + 1;
	libraryNameLength := libraryNameLength + 1;
    od;
    /* first, add a bss hunk with space for the library base pointer */
    addLong(lo, HUNK_UNIT);
    addLong(lo, 0);
    addLong(lo, HUNK_BSS);
    addLong(lo, 4 / 4);
    /* add an EXT hunk to define the base name as in the bss hunk */
    addLong(lo, HUNK_EXT);
    addName(lo, &BaseName[0], EXT_DEF);
    addLong(lo, 0);
    addLong(lo, 0);
    addLong(lo, HUNK_END);
    /* chuck out the code for the Open routine */
    addLong(lo, HUNK_UNIT);
    addLong(lo, 0);
    addLong(lo, HUNK_CODE);
    addLong(lo, (libraryNameLength + 3) / 4 + 32 / 4);
    addWord(lo, OP_MOVEL | 1 << 9 | M_ADIR << 6 | M_INC << 3 | R_SP);
    addWord(lo, OP_MOVEL | 0 << 9 | M_DDIR << 3 | M_INC << 3 | R_SP);
    addWord(lo, OP_MOVEL | R_SP << 9 | M_DEC << 6 | M_ADIR << 3 | 1);
    addWord(lo, OP_LEA | 1 << 9 | M_SPECIAL << 3 | M_PCDISP);
    addWord(lo, 24);
    addWord(lo, OP_MOVEL | R_SP << 9 | M_DEC << 6 | M_ADIR << 3 | R_LIB);
    addWord(lo,
	    OP_MOVEL | R_LIB << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABSLONG);
    addLong(lo, ABS_EXEC_BASE);
    addWord(lo, OP_JSR | M_DISP << 3 | R_LIB);
    addWord(lo, pretend(LVO_OPEN_LIBRARY, uint));
    addWord(lo, OP_MOVEL | R_LIB << 9 | M_ADIR << 6 | M_INC << 3 | R_SP);
    addWord(lo, OP_MOVEL | M_ABSLONG << 9 | M_SPECIAL << 6 | M_DDIR << 3 | 0);
    addLong(lo, 0);	/* reference the long in the above bss */
    addWord(lo, OP_RTS);
    /* add in the name of the library to open */
    addString(lo, libraryName);
    addString(lo, ".library");
    addByte(lo, 0);
    /* pad it to a long word boundary */
    if libraryNameLength & 0x2 ~= 0 then
	addByte(lo, 0);
	if libraryNameLength & 0x1 = 0 then
	    addByte(lo, 0);
	fi;
    elif libraryNameLength & 0x1 ~= 0 then
	addByte(lo, 0);
	addWord(lo, 0);
    fi;
    /* capitalize the first letter of the library name */
    if libraryName* >= 'a' and libraryName* <= 'z' then
	libraryName* := libraryName* - 32;
    fi;
    /* build the name of the Open routine */
    buildName(&entryName[0], "Open", libraryName, "Library");
    /* define the entry point and reference the base variable */
    addLong(lo, HUNK_EXT);
    addName(lo, &entryName[0], EXT_DEF);
    addLong(lo, 0);
    addName(lo, &BaseName[0], EXT_REF32);
    addLong(lo, 1);
    addLong(lo, 26);
    addLong(lo, 0);
    addLong(lo, HUNK_END);
    /* chuck out the code for the Close routine */
    addLong(lo, HUNK_UNIT);
    addLong(lo, 0);
    addLong(lo, HUNK_CODE);
    addLong(lo, (22 + 2) / 4);
    addWord(lo, OP_MOVEL | 1 << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABSLONG);
    addLong(lo, 0);	/* reference the long in the above bss */
    addWord(lo, OP_MOVEL | R_SP << 9 | M_DEC << 6 | M_ADIR << 3 | R_LIB);
    addWord(lo,
	    OP_MOVEL | R_LIB << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABSLONG);
    addLong(lo, ABS_EXEC_BASE);
    addWord(lo, OP_JSR | M_DISP << 3 | R_LIB);
    addWord(lo, pretend(LVO_CLOSE_LIBRARY, uint));
    addWord(lo, OP_MOVEL | R_LIB << 9 | M_ADIR << 6 | M_INC << 3 | R_SP);
    addWord(lo, OP_RTS);
    /* pad to a long word boundary */
    addWord(lo, 0);
    /* build the name of the Close routine */
    buildName(&entryName[0], "Close", libraryName, "Library");
    /* define the entry point and reference the base variable */
    addLong(lo, HUNK_EXT);
    addName(lo, &entryName[0], EXT_DEF);
    addLong(lo, 0);
    addName(lo, &BaseName[0], EXT_REF32);
    addLong(lo, 1);
    addLong(lo, 2);
    addLong(lo, 0);
    addLong(lo, HUNK_END);
    if DEBUG then
	writeString("leaving produceOpenClose\n");
    fi;
corp;

/*
 * generateCode -
 *	Generate the AmigaDOS object code for an interface routine for the
 *	passed entry description.
 */

proc generateCode(int offset; uint parCount; [16] uint regList;
	uint saveCount; uint saveMask, restoreMask; bool gotFloat)uint:
    register *Output_t lo;
    register uint i;
    uint baseRef;

    if DEBUG then
	writeString("generateCode\n");
    fi;
    lo := &LibOutput;
    if saveCount = 1 then
	addWord(lo, OP_MOVEL | R_SP << 9 | M_DEC << 6 | M_ADIR << 3 | R_LIB);
    else
	addWord(lo, OP_MOVEML | MOVEM_SAVE | M_DEC << 3 | R_SP);
	addWord(lo, saveMask);
    fi;
    for i from 1 upto parCount do
	addWord(lo, OP_MOVEL | make(
	    if gotFloat and regList[i - 1] < 8 then
		(regList[i - 1] & 7) >< 1
	    else
		regList[i - 1] & 7
	    fi, uint) << 9 |
	    if regList[i - 1] >= 8 then M_ADIR << 6 else M_DDIR << 6 fi |
	    M_DISP << 3 | R_SP);
	addWord(lo, (parCount - (i - 1) + saveCount) * 4);
    od;
    addWord(lo,
	    OP_MOVEL | R_LIB << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABSLONG);
    baseRef := CodePos;
    addLong(lo, 0);
    addWord(lo, OP_JSR | M_DISP << 3 | R_LIB);
    addWord(lo, offset);
    if saveCount = 1 then
	addWord(lo, OP_MOVEL | R_LIB << 9 | M_ADIR << 6 | M_INC << 3 | R_SP);
    else
	addWord(lo, OP_MOVEML | MOVEM_RESTORE | M_INC << 3 | R_SP);
	addWord(lo, restoreMask);
    fi;
    if parCount = 0 then
	addWord(lo, OP_RTS);
    else
	addWord(lo, OP_MOVEL | 1 << 9 | M_ADIR << 6 | M_INC << 3 | R_SP);
	if parCount <= 2 then
	    addWord(lo, OP_ADDQ | make((parCount * 4) & 7, uint) << 9 |
			S_LONG << 6 | M_ADIR << 3 | R_SP);
	else
	    addWord(lo, OP_ADD | R_SP << 9 | OM_WORD_ADDR << 6 |
				 M_SPECIAL << 3 | M_IMM);
	    addWord(lo, parCount * 4);
	fi;
	addWord(lo, OP_JMP | M_INDIR << 3 | 1);
    fi;
    /* round up to a long word boundary */
    if CodePos & 2 ~= 0 then
	addWord(lo, 0);
    fi;
    if DEBUG then
	writeString("leaving generateCode\n");
    fi;
    baseRef
corp;

/*
 * produceEntry -
 *	Generate the AmigaDOS object files for the interface routines and
 *	the LVO symbols from the passed entry description.
 */

proc produceEntry(int offset; uint parCount; [16] uint regList;
		  bool gotFloat)void:
    [NAME_SIZE] char LVOName;
    register *Output_t lo;
    register uint i, reg;
    uint saveCount, saveMask, restoreMask, baseRef, codeSize;

    if DEBUG then
	writeString("produceEntry\n");
    fi;

    lo := &LibOutput;
    if MakeLib then
	/* calculate MOVEM save and restore masks for the used regs */
	saveMask := 0b0000000000000010;
	restoreMask := 0b0100000000000000;
	saveCount := 1;
	for i from 1 upto parCount do
	    reg := regList[i - 1];
	    if reg % 8 >= 2 then
		saveCount := saveCount + 1;
		if saveMask & (1 << (15 - reg)) ~= 0 then
		    syntax("Duplicate register\n");
		fi;
		saveMask := saveMask | (1 << (15 - reg));
		restoreMask := restoreMask | (1 << reg);
	    fi;
	od;
	/* pass one of code generation - find out how big it is */
	PassOne := true;
	CodePos := 0;
	baseRef := generateCode(offset, parCount, regList,
				saveCount, saveMask, restoreMask, gotFloat);
	codeSize := CodePos;
	PassOne := false;
	/* start the actual program unit for the entry */
	addLong(lo, HUNK_UNIT);
	addLong(lo, 0); 	/* don't give the unit a name */
	addLong(lo, HUNK_CODE);
	addLong(lo, (codeSize + 3) / 4);
	/* pass two of code generation - add the code to the text hunk */
	CodePos := 0;
	baseRef := generateCode(offset, parCount, regList,
				saveCount, saveMask, restoreMask, gotFloat);
	/* close off with the external refs/defs */
	addLong(lo, HUNK_EXT);
	addName(lo, &EntryName[0], EXT_DEF);
	addLong(lo, 0);
	addName(lo, &BaseName[0], EXT_REF32);
	addLong(lo, 1);
	addLong(lo, baseRef);
	addLong(lo, 0);
	addLong(lo, HUNK_END);
    fi;

    if MakeLVO then
	/* add a simple definition to the LVO file */
	buildName(&LVOName[0], "LVO_", &EntryName[0], "");
	addName(&LVOOutput, &LVOName[0], EXT_ABS);
	addLong(&LVOOutput, offset);
    fi;

    if DEBUG then
	writeString("leaving produceEntry\n");
    fi;
corp;

/*
 * findSymAddress -
 *	Find the actual ROM address of the given symbol and add it to the
 *	output file.
 */

proc findSymAddress(int offset)void:
    [NAME_SIZE] char symName;
    register *uint p;

    p := pretend(LibPtr, *uint) + offset;
    if p* = OP_JMP | M_SPECIAL << 3 | M_ABSLONG then
	if MakeSym then
	    buildName(&symName[0], "\#", &EntryName[0], "");
	    addName(&SymOutput, &symName[0], EXT_ABS);
	    addWord(&SymOutput, (p + 2)*);
	    addWord(&SymOutput, (p + 4)*);
	fi;
	if PrintSym then
	    writeString(&EntryName[0]);
	    writeString(" @ ");
	    writeHex((p + 2)*);
	    writeHex((p + 4)*);
	    writeString("\n");
	fi;
    else
	writeString(&EntryName[0]);
	writeString(" - not a JMP ABS\n");
    fi;
corp;

/*
 * printSpecials - special code to do predefined entries for PrintSym.
 */

proc printSpecials()void:

    CharsCopy(&EntryName[0], "OPEN");
    findSymAddress(LIB_OPEN);
    CharsCopy(&EntryName[0], "CLOSE");
    findSymAddress(LIB_CLOSE);
    CharsCopy(&EntryName[0], "EXPUNGE");
    findSymAddress(LIB_EXPUNGE);
    CharsCopy(&EntryName[0], "EXTFUNC");
    findSymAddress(LIB_EXTFUNC);
    if IsDevice then
	CharsCopy(&EntryName[0], "BEGINIO");
	findSymAddress(DEV_BEGINIO);
	CharsCopy(&EntryName[0], "ABORTIO");
	findSymAddress(DEV_ABORTIO);
    fi;
corp;

/*
 * processEntry -
 *	Handle an entry definition with the passed offset.
 */

proc processEntry(int offset)void:
    register uint parCount, i, reg;
    register *FuncList_t fl;
    [16] uint regList;
    bool gotFloat;

    getName();
    EntryName := InputName;
    if MakeHead then
	fl := AllocMem(sizeof(FuncList_t), 0);
	if fl = nil then
	    abort("Can't allocate space for function name save\n", 30);
	fi;
	fl*.fl_next := nil;
	fl*.fl_name := EntryName;
	FuncTail* := fl;
	FuncTail := &fl*.fl_next;
	FuncCount := FuncCount + 1;
    fi;
    if DEBUG then
	writeString("processing entry '");
	writeString(&EntryName[0]);
	writeString("'\n");
    fi;
    if InputChar ~= '(' then
	syntax("No opening '(' for parameter name list\n");
    fi;
    getChar();
    gotFloat := false;
    parCount := 0;
    while InputChar ~= ')' do
	parCount := parCount + 1;
	getName();
	if InputChar = '.' then
	    getChar();
	    if InputChar = 'f' then
		parCount := parCount + 1;
		gotFloat := true;
	    else
		syntax("Unknown length letter on parameter name\n");
	    fi;
	    getChar();
	fi;
	if InputChar = ',' then
	    getChar();
	    while InputChar = ' ' do
		getChar();
	    od;
	elif InputChar ~= ')' then
	    syntax("Bad separator in parameter name list\n");
	fi;
    od;
    getChar();
    if InputChar ~= '(' and parCount ~= 0 then
	syntax("Missing '(' for register list\n");
    fi;
    if InputChar = '(' then
	getChar();
	i := 0;
	while InputChar ~= ')' do
	    if InputChar = 'A' or InputChar = 'a' then
		reg := 8;
	    elif InputChar = 'D' or InputChar = 'd' then
		reg := 0;
	    else
		syntax("Bad register name letter\n");
	    fi;
	    getChar();
	    if InputChar >= '0' and InputChar <= '7' then
		reg := reg + (InputChar - '0');
	    else
		syntax("Bad register name digit\n");
	    fi;
	    if reg >= 8 + 6 then
		/* sp or a6 */
		syntax("Can't use sp or a6\n");
	    fi;
	    regList[i] := reg;
	    getChar();
	    if InputChar = ',' or InputChar = '/' then
		getChar();
	    elif InputChar ~= ')' then
		syntax("Bad separator in register list\n");
	    fi;
	    i := i + 1;
	od;
	if i ~= parCount then
	    syntax("Register count different from parameter count\n");
	fi;
    fi;
    if PrintLib then
	writeString(&EntryName[0]);
	writeString(" @ -");
	writeNumber(- offset);
	writeString(" (hex -");
	writeHex(- offset);
	writeString(")");
	if gotFloat then
	    writeString(" (all float pars)");
	fi;
	writeString("\n");
    fi;
    if MakeLib or MakeLVO then
	produceEntry(offset, parCount, regList, gotFloat);
    fi;
    if MakeSym or PrintSym then
	findSymAddress(offset);
    fi;
    if DEBUG then
	writeString("processEntry done\n");
    fi;
corp;

/*
 * isName -
 *	Compare the current input name with the passed one. Return 'true' if
 *	they are the same.
 */

proc isName(register *char subject)bool:
    register *char p;

    p := &InputName[0];
    while p* = subject* and p* ~= '\e' do
	p := p + 1;
	subject := subject + 1;
    od;
    p* = subject*
corp;

/*
 * buildHeader - build the library header file.
 */

proc buildHeader(*char fullName; uint nameLen)void:
    uint R_SAVE = 5;
    [25] char timeBuffer;
    register *Output_t ho;
    register *FuncList_t fl;
    register ulong offset;
    Library_t fakeLib;			/* used to get offsets */

    ho := &HeadOutput;
    fl := FuncList;

    /* define a code hunk for the library file */
    addLong(ho, HUNK_UNIT);
    addLong(ho, 0);			/* no name */
    addLong(ho, HUNK_CODE);
    addLong(ho,
	(				/* length of the code hunk */
	 28 +				/* the romtag */
	 16 +				/* the library init table */
	 28 +				/* the data init table */
	 16 +				/* the init routine */
	 14 +				/* the open routine */
	 20 +				/* the close routine */
	 66 +				/* the expunge routine */
	 04 +				/* the null routine */
	 (FuncCount + 5) * 4 +		/* the function table */
	 (nameLen + 9 + 		/* the library name */
	  nameLen + 21 +		/* the library identifier */
	  3) / 4 * 4			/* round up to nearest longword */
	) / 4); 			/* want length in longwords */

    /* add the romtag structure for the library (offset = 0) */
    addWord(ho, RTC_MATCHWORD);
    addLong(ho, 0);			/* ref to self */
    addLong(ho, 26);			/* ref to end of romtag */
    addByte(ho, RTF_AUTOINIT);		/* INIT is table, not func */
    addByte(ho, (Version1 - '0') * 10 + (Version2 - '0'));	/* version */
    addByte(ho, NT_LIBRARY);
    addByte(ho, 0);			/* priority 0 */
    addLong(ho, 212 + FuncCount * 4);	/* ref to lib name */
    addLong(ho, 221 + FuncCount * 4 + nameLen);/* ref to lib identifier */
    addLong(ho, 28);			/* ref to lib init table */
    addWord(ho, 0);			/* pad to longword boundary */

    /* add the library init table (offset = 28) */
    addLong(ho, sizeof(Library_t) + 8); /* positive is ExecBase + my seglist */
    addLong(ho, 192);			/* ref to func table */
    addLong(ho, 44);			/* ref to data init */
    addLong(ho, 72);			/* ref to init routine */

    /* add the data initialization structure (offset = 44) */
    addByte(ho, 0b10100000);
    addByte(ho, &fakeLib.lib_Node.ln_Type - pretend(&fakeLib, *byte));
    addByte(ho, NT_LIBRARY);
    addByte(ho, 0);			/* pad to word boundary */
    addByte(ho, 0b10000000);
    addByte(ho, &fakeLib.lib_Node.ln_Name - pretend(&fakeLib, **char));
    addLong(ho, 212 + FuncCount * 4);	/* ref to library name */
    addByte(ho, 0b10100000);
    addByte(ho, &fakeLib.lib_Flags - pretend(&fakeLib, *ushort));
    addByte(ho, LIBF_CHANGED | LIBF_SUMUSED);
    addByte(ho, 0);			/* pad to word boundary */
    addByte(ho, 0b10010000);
    addByte(ho, &fakeLib.lib_Version - pretend(&fakeLib, *uint));
    addWord(ho, make(Version1 - '0', uint) * 10 + (Version2 - '0'));
/*
    addWord(ho, 1);
*/
    addByte(ho, 0b10000000);
    addByte(ho, &fakeLib.lib_IdString - pretend(&fakeLib, **char));
    addLong(ho, 221 + FuncCount * 4 + nameLen);/* ref to id string */
    addLong(ho, 0);			/* end of data init table */

    /* add the library init routine (offset = 72) */
    /* save A1 so we can use it here */
    addWord(ho, OP_MOVEL | R_SP << 9 | M_DEC << 6 | M_ADIR << 3 | 1);
    /* get ptr to my Library_t from D0 to A1 */
    addWord(ho, OP_MOVEL | 1 << 9 | M_ADIR << 6 | M_DDIR << 3 | 0);
    /* save our SegList ptr into offset 4 past the Library_t */
    addWord(ho, OP_MOVEL | 1 << 9 | M_DISP << 6 | M_ADIR << 3 | 0);
    addWord(ho, sizeof(Library_t) + 4);
    /* save a copy of ExecBase into offset 0 past the Library_t */
    addWord(ho, OP_MOVEL | 1 << 9 | M_DISP << 6 | M_ADIR << 3 | 6);
    addWord(ho, sizeof(Library_t));
    /* restore A1 */
    addWord(ho, OP_MOVEL | 1 << 9 | M_ADIR << 6 | M_INC << 3 | R_SP);
    /* and return to Exec */
    addWord(ho, OP_RTS);

    /* add the library open routine (offset = 88) */
    addWord(ho, OP_ADDQ | 1 << 9 | S_WORD << 6 | M_DISP << 3 | R_LIB);
    addWord(ho, &fakeLib.lib_OpenCnt - pretend(&fakeLib, *uint));
    addWord(ho, OP_BCLR | M_DISP << 3 | R_LIB);
    addWord(ho, 3);
    addWord(ho, &fakeLib.lib_Flags - pretend(&fakeLib, *ushort));
    addWord(ho, OP_MOVEL | 0 << 9 | M_DDIR << 6 | M_ADIR << 3 | R_LIB);
    addWord(ho, OP_RTS);

    /* add the library close routine (offset = 102) */
    addWord(ho, OP_CLR | S_LONG << 6 | M_DDIR << 3 | 0);
    addWord(ho, OP_SUBQ | 1 << 9 | S_WORD << 6 | M_DISP << 3 | R_LIB);
    addWord(ho, &fakeLib.lib_OpenCnt - pretend(&fakeLib, *uint));
    addWord(ho, OP_BCC | CC_NE << 8 | 10);
    addWord(ho, OP_BTST | M_DISP << 3 | R_LIB);
    addWord(ho, 3);
    addWord(ho, &fakeLib.lib_Flags - pretend(&fakeLib, *ushort));
    addWord(ho, OP_BCC | CC_EQ << 8 | 2);
    addWord(ho, OP_BSR | 2);
    addWord(ho, OP_RTS);

    /* add the library expunge routine (offset = 122) */
    addWord(ho, OP_MOVEML | MOVEM_SAVE | M_DEC << 3 | R_SP);
    addWord(ho, 0b0000000001000110);
    addWord(ho, OP_CLR | S_LONG << 6 | M_DDIR << 3 | 0);
    addWord(ho, OP_MOVEL | R_SAVE << 9 | M_ADIR << 6 | M_ADIR << 3 | R_LIB);
    addWord(ho, OP_TST | S_WORD << 6 | M_DISP << 3 | R_SAVE);
    addWord(ho, &fakeLib.lib_OpenCnt - pretend(&fakeLib, *uint));
    addWord(ho, OP_BCC | CC_EQ << 8 | 8);
    addWord(ho, OP_BSET | M_DISP << 3 | R_SAVE);
    addWord(ho, 3);
    addWord(ho, &fakeLib.lib_Flags - pretend(&fakeLib, *ushort));
    addWord(ho, OP_BCC | CC_T << 8 | 38);
    addWord(ho, OP_MOVEL | 1 << 9 | M_ADIR << 6 | M_ADIR << 3 | R_SAVE);
    addWord(ho,
	    OP_MOVEL | R_LIB << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABSLONG);
    addLong(ho, ABS_EXEC_BASE);
    addWord(ho, OP_JSR | M_DISP << 3 | R_LIB);
    addWord(ho, pretend(LVO_REMOVE, uint));
    addWord(ho, OP_MOVEL | 1 << 9 | M_ADIR << 6 | M_ADIR << 3 | R_SAVE);
    addWord(ho, OP_MOVEW | 0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_SAVE);
    addWord(ho, &fakeLib.lib_NegSize - pretend(&fakeLib, *uint));
    addWord(ho, OP_SUB | 1 << 9 | OM_WORD_ADDR << 6 | M_DDIR << 3 | 0);
    addWord(ho, OP_ADD | 0 << 9 | (OM_TO_REG | S_WORD) << 6 |
			 M_DISP << 3 | R_SAVE);
    addWord(ho, &fakeLib.lib_PosSize - pretend(&fakeLib, *uint));
    addWord(ho,
	    OP_MOVEL | R_LIB << 9 | M_ADIR << 6 | M_SPECIAL << 3 | M_ABSLONG);
    addLong(ho, ABS_EXEC_BASE);
    addWord(ho, OP_JSR | M_DISP << 3 | R_LIB);
    addWord(ho, pretend(LVO_FREE_MEM, uint));
    addWord(ho, OP_MOVEL | 0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_SAVE);
    addWord(ho, sizeof(Library_t) + 4);
    addWord(ho, OP_MOVEML | MOVEM_RESTORE | M_INC << 3 | R_SP);
    addWord(ho, 0b0110001000000000);
    addWord(ho, OP_RTS);

    /* add the library null routine (offset = 188) */
    addWord(ho, OP_CLR | S_LONG << 6 | M_DDIR << 3 | 0);
    addWord(ho, OP_RTS);

    /* add the function table (offset = 192) */
    addLong(ho, 88);			/* ref to library open routine */
    addLong(ho, 102);			/* ref to library close routine */
    addLong(ho, 122);			/* ref to library expunge routine */
    addLong(ho, 188);			/* ref to library null routine */
    fl := FuncList;
    while fl ~= nil do
	addLong(ho, 0); 		/* ref to library function */
	fl := fl*.fl_next;
    od;
    addLong(ho, -1);			/* end of function table */

    /* add the name of the library (offset = 212 + FuncCount * 4) */
    addString(ho, fullName);
    addByte(ho, 0);

    /* add the library identification (offs = 221 + FuncCount * 4 + nameLen) */
    ConvTime(GetCurrentTime(), &timeBuffer[0]);
    /* resulting format:
       Sun May 14 12:32:10 1989
     */
    (fullName + nameLen)* := '\e';
    addString(ho, fullName);
    addByte(ho, ' ' - '\e');
    addByte(ho, Version1 - '\e');
    addByte(ho, '.' - '\e');
    addByte(ho, Version2 - '\e');
    addString(ho, " (");
    addByte(ho, timeBuffer[8] - '\e');
    addByte(ho, timeBuffer[9] - '\e');
    addByte(ho, ' ' - '\e');
    timeBuffer[8] := '\e';
    addString(ho, &timeBuffer[4]);
    addString(ho, &timeBuffer[20]);
    addString(ho, ")\r\n");
    addByte(ho, 0);
    if nameLen & 1 = 0 then
	addWord(ho, 0); 		/* pad to longword boundary */
    fi;

    /* the code is done - add a reloc32 hunk to fix pointers within it */
    addLong(ho, HUNK_RELOC32);
    addLong(ho, 14);			/* there are 14 references */
    addLong(ho, 0);			/* ref to hunk #0 */
    addLong(ho, 2);			/* romtag => self */
    addLong(ho, 6);			/* romtag => end of romtag */
    addLong(ho, 14);			/* romtag => lib name */
    addLong(ho, 18);			/* romtag => lib ident */
    addLong(ho, 22);			/* romtag => lib init table */
    addLong(ho, 32);			/* lib init table => func table */
    addLong(ho, 36);			/* lib init table => data init table */
    addLong(ho, 40);			/* lib init table => init routine */
    addLong(ho, 50);			/* data init table => lib name */
    addLong(ho, 64);			/* data init table => lib ident */
    addLong(ho, 192);			/* func table => open routine */
    addLong(ho, 196);			/* func table => close routine */
    addLong(ho, 200);			/* func table => expunge routine */
    addLong(ho, 204);			/* func table => null routine */
    addLong(ho, 0);			/* end of relocs */

    /* add a HUNK_EXT to refer to the actual library functions */
    addLong(ho, HUNK_EXT);
    fl := FuncList;
    offset := 192 + 16;
    while fl ~= nil do
	addName(&HeadOutput, &fl*.fl_name[0], EXT_REF32);
	addLong(&HeadOutput, 1);
	addLong(&HeadOutput, offset);
	offset := offset + 4;
	fl := fl*.fl_next;
    od;
    addLong(ho, 0);			/* end of hunk_ext */

    addLong(ho, HUNK_END);
corp;

/*
 * fixName - get the full name of the library/device and maybe open it.
 */

proc fixName(*char libraryName, buff; *uint pLen)bool:
    register *char p, q;
    register uint len;
    bool ok;

    p := libraryName;
    q := buff;
    len := 0;
    while p* ~= '\e' do
	q* := p*;
	q := q + 1;
	p := p + 1;
	len := len + 1;
    od;
    if IsDevice then
	p := ".device";
    else
	p := ".library";
    fi;
    while p* ~= '\e' do
	q* := p*;
	q := q + 1;
	p := p + 1;
    od;
    q* := '\e';
    pLen* := len;
    ok := true;
    if MakeSym or PrintSym then
	if DEBUG then
	    writeString("Trying to open library/device '");
	    writeString(buff);
	    writeString("'\n");
	fi;
	if IsDevice then
	    Req := AllocMem(200, MEMF_PUBLIC | MEMF_CLEAR);
	    if Req = nil then
		writeString("No memory for request on '");
		ok := false;
	    elif OpenDevice(buff, 0, Req, 0) ~= 0 then
		writeString("Can't open device '");
		ok := false;
	    else
		LibPtr := &Req*.io_Device*.dd_Library;
	    fi;
	else
	    LibPtr := OpenLibrary(buff, 0);
	    if LibPtr = nil then
		writeString("Can't open library '");
		ok := false;
	    fi;
	fi;
	if not ok then
	    writeString(buff);
	    writeString("' - syms cancelled\n");
	    Close(SymOutput.o_fd);
	    SymOutput.o_fd := 0;
	    MakeSym := false;
	    PrintSym := false;
	else
	    if MakeSym then
		addLong(&SymOutput, HUNK_UNIT);
		addLong(&SymOutput, 0);
		addLong(&SymOutput, HUNK_EXT);
	    fi;
	fi;
    fi;
    ok
corp;

/*
 * parseFile -
 *	Convert the given .fd file to the given .o files. The library name is
 *	passed to us.
 */

proc parseFile(*char libraryName)void:
    [50] char fullName;
    int offset;
    uint len;
    char version1;
    bool gotBase, gotBias, gotEnd, skipping, libDevOpen, ok;

    if DEBUG then
	writeString("parsing file\n");
    fi;
    fullName[0] := '\e';
    if MakeLVO then
	/* start the LVO file up with a HUNK_UNIT, no name and a HUNK_EXT */
	addLong(&LVOOutput, HUNK_UNIT);
	addLong(&LVOOutput, 0);
	addLong(&LVOOutput, HUNK_EXT);
    fi;
    /* now go parse and handle the descriptions for the library entries */
    FuncList := nil;
    offset := -30;
    gotBase := false;
    gotBias := false;
    gotEnd := false;
    skipping := false;
    while not Eof and not gotEnd do
	if InputChar = '*' then
	    /* a comment, skip it */
	    if DEBUG then
		writeString("skipping a comment\n");
	    fi;
	elif InputChar = '\#' then
	    /* hopefully a special 'command' line */
	    getChar();
	    if InputChar ~= '\#' then
		syntax("Line with single \#\n");
	    fi;
	    /* yep, a command line */
	    getChar();
	    getName();
	    if isName("public") then
		skipping := false;
		if DEBUG then
		    writeString("got a 'public'\n");
		fi;
	    elif isName("private") then
		skipping := true;
		if DEBUG then
		    writeString("got a 'private'\n");
		fi;
	    elif isName("base") then
		if DEBUG then
		    writeString("got 'base'\n");
		fi;
		if gotBase then
		    syntax("Multiple 'base' lines.\n");
		fi;
		/* save the base name for external referencing */
		getChar();
		getName();
		BaseName := InputName;
		gotBase := true;
	    elif isName("bias") then
		if DEBUG then
		    writeString("got 'bias'\n");
		fi;
		while InputChar = ' ' do
		    getChar();
		od;
		offset := 0;
		while InputChar >= '0' and InputChar <= '9' do
		    offset := offset * 10 - (InputChar - '0');
		    getChar();
		od;
		if not gotBase then
		    syntax("No base when get bias\n");
		fi;
		if offset = -42 and not gotBias then
		    IsDevice := true;
		fi;
		if not gotBias then
		    libDevOpen := fixName(libraryName, &fullName[0], &len);
		    if MakeLib then
			if IsDevice then
			    /* this is a special 'device', not a library. */
			    produceSetDevice(libraryName);
			else
			    /* now start the lib file up with the
			       OpenXXXLibrary and CloseXXXLibrary entry
			       points. */
			    produceOpenClose(libraryName);
			fi;
		    fi;
		    if PrintSym and PrintAll then
			printSpecials();
		    fi;
		    gotBias := true;
		fi;
	    elif isName("end") then
		/* 'end' - end of the .fd file */
		if DEBUG then
		    writeString("got 'end' line\n");
		fi;
		gotEnd := true;
	    elif isName("version") then
		if DEBUG then
		    writeString("got 'version'\n");
		fi;
		while InputChar = ' ' do
		    getChar();
		od;
		ok := false;
		if InputChar >= '0' and InputChar <= '9' then
		    version1 := InputChar;
		    getChar();
		    if InputChar = '.' then
			getChar();
			if InputChar >= '0' and InputChar <= '9' then
			    Version1 := version1;
			    Version2 := InputChar;
			    ok := true;
			fi;
		    fi;
		fi;
		if not ok then
		    syntax("Illegal 'version' format\n");
		fi;
	    else
		syntax("Unknown \#\# command\n");
	    fi;
	elif InputChar >= 'A' and InputChar <= 'Z' or
		InputChar >= 'a' and InputChar <= 'z' then
	    /* a function name! */
	    if not gotBias then
		syntax("Function name before any 'bias'.\n");
	    fi;
	    if not skipping then
		processEntry(offset);
	    fi;
	    offset := offset - 6;	/* hey! a magic number!! */
	else
	    syntax("Unrecognized input line\n");
	fi;
	while InputChar ~= '\n' do
	    getChar();
	od;
	getChar();
	CurrentLine := CurrentLine + 1;
    od;
    if MakeLVO then
	/* close the LVO file off with a 0 to end EXT, and a HUNK_END */
	addLong(&LVOOutput, 0);
	addLong(&LVOOutput, HUNK_END);
    fi;
    if MakeSym or PrintSym then
	if IsDevice then
	    CloseDevice(Req);
	    FreeMem(Req, 200);
	else
	    CloseLibrary(LibPtr);
	fi;
	if MakeSym then
	    addLong(&SymOutput, 0);
	    addLong(&SymOutput, HUNK_END);
	fi;
    fi;
    if MakeHead then
	buildHeader(&fullName[0], len);
    fi;
    if DEBUG then
	writeString("parse done\n");
    fi;
corp;

/*
 * setupFile - set up one of the output files.
 */

proc setupFile(register *Output_t o; *char s1, s2, s3)void:
    [100] char fileName;

    /* build the lib file name */
    buildName(&fileName[0], s1, s2, s3);
    /* next, attempt to open/create the lib .o file */
    DeleteFile(&fileName[0]);
    o*.o_fd := Open(&fileName[0], MODE_NEWFILE);
    if o*.o_fd = 0 then
	writeString("Can't open/create output file '");
	writeString(&fileName[0]);
	writeString("'\n");
	closeAll();
	Exit(20);
    fi;
    o*.o_pos := 0;
corp;

/*
 * processFile -
 *	Process an input file of the given name.
 */

proc processFile(*char fdFileName)void:
    register *char base, end;

    if DEBUG then
	writeString("processing file '");
	writeString(fdFileName);
	writeString("'\n");
    fi;
    /* first, attempt to open the .fd file */
    FdFd := Open(fdFileName, MODE_READONLY);
    if FdFd = 0 then
	writeString("Can't open input file '");
	writeString(fdFileName);
	writeString("'\n");
	Exit(20);
    fi;
    IsDevice := false;
    /* next, find the basic file name (strip off any leading path) */
    base := fdFileName;
    while base* ~= '\e' do
	base := base + 1;
    od;
    while base ~= fdFileName and base* ~= '/' do
	base := base - 1;
    od;
    if base* = '/' then
	base := base + 1;
    fi;
    /* also strip off any trailing portion (stick to just letters) */
    end := base;
    while end* >= 'a' and end* <= 'z' or end* >= 'A' and end* <= 'Z' do
	end := end + 1;
    od;
    if end = base then
	writeString("File name doesn't start with library name. '");
	writeString(fdFileName);
	writeString("'\n");
	Close(FdFd);
	Exit(20);
    fi;
    end* := '\e';
    if DEBUG then
	writeString("base name = '");
	writeString(base);
	writeString("'\n");
    fi;
    LibOutput.o_fd := 0;
    LVOOutput.o_fd := 0;
    SymOutput.o_fd := 0;
    HeadOutput.o_fd := 0;
    FuncList := nil;
    FuncTail := &FuncList;
    FuncCount := 0;
    if MakeLib then
	setupFile(&LibOutput, base, ".o", "");
    fi;
    if MakeLVO then
	setupFile(&LVOOutput, "LVO_", base, ".o");
    fi;
    if MakeSym then
	setupFile(&SymOutput, base, "_sym.o", "");
    fi;
    if MakeHead then
	setupFile(&HeadOutput, base, "_head.o", "");
    fi;
    /* initialize file buffer, then go process the file */
    Eof := false;
    FdMax := 0;
    FdPos := 0;
    CurrentFile := base;
    CurrentLine := 1;
    getChar();
    parseFile(base);
    /* flush everything, close files and return */
    closeAll();
corp;

/*
 * main -
 *	The main program. Get the command line parameters, parse them, and
 *	pass the two file names to 'process'.
 */

proc main()void:
    extern GetPars(*ulong pLen; **char pPtr)void;
    register *char parPtr;
    *char fdFileName, xParPtr;
    register ulong parLen;
    ulong xParLen;
    [2] char buff;
    bool hadFile;

    if OpenExecLibrary(0L0) ~= nil then
	if OpenDosLibrary(0L0) ~= nil then
	    Stdout := Output();
	    /* retrieve the D0/A0 values saved by our startup code */
	    GetPars(&xParLen, &xParPtr);
	    parLen := xParLen;
	    parPtr := xParPtr;
	    MakeLib := false;
	    PrintLib := false;
	    MakeLVO := false;
	    MakeSym := false;
	    PrintSym := false;
	    PrintAll := false;
	    MakeHead := false;
	    Version1 := '1';
	    Version2 := '0';
	    /* loop through all of the .fd file names given */
	    hadFile := false;
	    while
		/* skip leading/separating blanks or trailing newline */
		while parLen ~= 0 and
			(parPtr* = ' ' or parPtr* = '\n' or parPtr* = '\r') do
		    parPtr := parPtr + 1;
		    parLen := parLen - 1;
		od;
		parLen ~= 0
	    do
		if parPtr* = '-' then
		    /* a parameter, not a file name */
		    while
			parPtr := parPtr + 1;
			parLen := parLen - 1;
			parLen ~= 0 and parPtr* ~= ' ' and
			    parPtr* ~= '\t' and parPtr* ~= '\n' and
			    parPtr* ~= '\r'
		    do
			case parPtr*
			incase 'l':
			    MakeLib := true;
			incase 'L':
			    PrintLib := true;
			incase 'o':
			    MakeLVO := true;
			incase 's':
			    MakeSym := true;
			incase 'S':
			    PrintSym := true;
			incase 'A':
			    PrintAll := true;
			incase 'd':
			    MakeHead := true;
			incase 'v':
			    parPtr := parPtr + 1;
			    parLen := parLen - 1;
			    if parLen < 3 or (parPtr + 1)* ~= '.' or
				not (parPtr* >= '0' and parPtr* <= '9' and
				     (parPtr + 2)* >= '0' and
				     (parPtr + 2)* <= '9')
			    then
				writeString("Illegal version: ");
				writeString(parPtr);
				writeString("\n");
				parLen := 0;
			    else
				Version1 := parPtr*;
				Version2 := (parPtr + 2)*;
			    fi;
			default:
			    writeString("Unknown flag '");
			    buff[0] := parPtr*;
			    buff[1] := '\e';
			    writeString(&buff[0]);
			    writeString("' - ignored\n");
			    writeString("Flags are:\n");
			    writeString("   l - produce normal lib file\n");
			    writeString("   L - print lib entries\n");
			    writeString("   o - produce LVO file\n");
			    writeString("   s - produce sym file\n");
			    writeString("   S - print sym value\n");
			    writeString("   A - print special values also\n");
			    writeString("   d - build disk library header\n");
			    writeString("   vN.M - set library version\n");
			esac;
		    od;
		else
		    hadFile := true;
		    /* skip over the characters of the file name */
		    fdFileName := parPtr;
		    while parLen ~= 0 and parPtr* ~= ' ' and
			    parPtr* ~= '\n' and parPtr* ~= '\r' do
			parPtr := parPtr + 1;
			parLen := parLen - 1;
		    od;
		    parPtr* := '\e';
		    if parLen ~= 0 then
			parPtr := parPtr + 1;
			parLen := parLen - 1;
		    fi;
		    writeString(fdFileName);
		    writeString(":\n");
		    processFile(fdFileName);
		fi;
	    od;
	    if not hadFile then
		writeString("Use is: fdCompile -lLosSAd fdfile ... fdfile\n");
	    fi;
	    CloseDosLibrary();
	fi;
	CloseExecLibrary();
    fi;
corp;
